@@ -8,7 +8,7 @@ require Exporter;
 use vars qw/@ISA $VERSION/;
 @ISA = qw(Exporter);
 
-$VERSION = '0.35';
+$VERSION = '0.37';
 
 # Package to store unsigned big integers in decimal and do math with them
 
@@ -125,30 +125,35 @@ BEGIN
   # determine how many digits fit into an integer and can be safely added 
   # together plus carry w/o causing an overflow
 
-  # this below detects 15 on a 64 bit system, because after that it becomes
-  # 1e16  and not 1000000 :/ I can make it detect 18, but then I get a lot of
-  # test failures. Ugh! (Tomake detect 18: uncomment lines marked with *)
   use integer;
-  my $bi = 5;			# approx. 16 bit
-  $num = int('9' x $bi);
-  # $num = 99999; # *
-  # while ( ($num+$num+1) eq '1' . '9' x $bi)	# *
-  while ( int($num+$num+1) eq '1' . '9' x $bi)
-    {
-    $bi++; $num = int('9' x $bi);
-    # $bi++; $num *= 10; $num += 9;	# *
-    }
-  $bi--;				# back off one step
+
+  ############################################################################
+  # the next block is no longer important
+
+  ## this below detects 15 on a 64 bit system, because after that it becomes
+  ## 1e16  and not 1000000 :/ I can make it detect 18, but then I get a lot of
+  ## test failures. Ugh! (Tomake detect 18: uncomment lines marked with *)
+
+  #my $bi = 5;			# approx. 16 bit
+  #$num = int('9' x $bi);
+  ## $num = 99999; # *
+  ## while ( ($num+$num+1) eq '1' . '9' x $bi)	# *
+  #while ( int($num+$num+1) eq '1' . '9' x $bi)
+  #  {
+  #  $bi++; $num = int('9' x $bi);
+  #  # $bi++; $num *= 10; $num += 9;	# *
+  #  }
+  #$bi--;				# back off one step
   # by setting them equal, we ignore the findings and use the default
   # one-size-fits-all approach from former versions
-  $bi = $e;				# XXX, this should work always
+  my $bi = $e;				# XXX, this should work always
 
   __PACKAGE__->_base_len($e,$bi);	# set and store
 
   # find out how many bits _and, _or and _xor can take (old default = 16)
   # I don't think anybody has yet 128 bit scalars, so let's play safe.
   local $^W = 0;	# don't warn about 'nonportable number'
-  $AND_BITS = 15; $XOR_BITS = 15; $OR_BITS  = 15;
+  $AND_BITS = 15; $XOR_BITS = 15; $OR_BITS = 15;
 
   # find max bits, we will not go higher than numberofbits that fit into $BASE
   # to make _and etc simpler (and faster for smaller, slower for large numbers)
@@ -221,6 +226,7 @@ sub _two
 
 sub _copy
   {
+  # make a true copy
   [ @{$_[1]} ];
   }
 
@@ -245,7 +251,7 @@ sub _str
   # leading zero parts in internal representation)
   $l --; $ret .= int($ar->[$l]); $l--;
   # Interestingly, the pre-padd method uses more time
-  # the old grep variant takes longer (14 to 10 sec)
+  # the old grep variant takes longer (14 vs. 10 sec)
   my $z = '0' x ($BASE_LEN-1);                            
   while ($l >= 0)
     {
@@ -291,8 +297,7 @@ sub _add
  
   # for each in Y, add Y to X and carry. If after that, something is left in
   # X, foreach in X add carry to X and then return X, carry
-  # Trades one "$j++" for having to shift arrays, $j could be made integer
-  # but this would impose a limit to number-length of 2**32.
+  # Trades one "$j++" for having to shift arrays
   my $i; my $car = 0; my $j = 0;
   for $i (@$y)
     {
@@ -309,8 +314,7 @@ sub _add
 sub _inc
   {
   # (ref to int_num_array, ref to int_num_array)
-  # routine to add 1 to a base 1eX numbers
-  # This routine modifies array x
+  # Add 1 to $x, modify $x in place
   my ($c,$x) = @_;
 
   for my $i (@$x)
@@ -325,17 +329,16 @@ sub _inc
 sub _dec
   {
   # (ref to int_num_array, ref to int_num_array)
-  # routine to add 1 to a base 1eX numbers
-  # This routine modifies array x
+  # Sub 1 from $x, modify $x in place
   my ($c,$x) = @_;
 
   my $MAX = $BASE-1;				# since MAX_VAL based on MBASE
   for my $i (@$x)
     {
     last if (($i -= 1) >= 0);			# early out
-    $i = $MAX;					# overflow, next
+    $i = $MAX;					# underflow, next
     }
-  pop @$x if $x->[-1] == 0 && @$x > 1;		# last overflowed (but leave 0)
+  pop @$x if $x->[-1] == 0 && @$x > 1;		# last underflowed (but leave 0)
   $x;
   }                                                                             
 
@@ -513,8 +516,18 @@ sub _div_use_mul
   {
   # ref to array, ref to array, modify first array and return remainder if 
   # in list context
-  my ($c,$x,$yorg) = @_;
 
+  # see comments in _div_use_div() for more explanations
+
+  my ($c,$x,$yorg) = @_;
+  
+  # the general div algorithmn here is about O(N*N) and thus quite slow, so
+  # we first check for some special cases and use shortcuts to handle them.
+
+  # This works, because we store the numbers in a chunked format where each
+  # element contains 5..7 digits (depending on system).
+
+  # if both numbers have only one element:
   if (@$x == 1 && @$yorg == 1)
     {
     # shortcut, $yorg and $x are two small numbers
@@ -530,6 +543,8 @@ sub _div_use_mul
       return $x; 
       }
     }
+
+  # if x has more than one, but y has only one element:
   if (@$yorg == 1)
     {
     my $rem;
@@ -549,6 +564,69 @@ sub _div_use_mul
     return $x;
     }
 
+  # now x and y have more than one element
+
+  # check whether y has more elements than x, if yet, the result will be 0
+  if (@$yorg > @$x)
+    {
+    my $rem;
+    $rem = [@$x] if wantarray;                  # make copy
+    splice (@$x,1);                             # keep ref to original array
+    $x->[0] = 0;                                # set to 0
+    return ($x,$rem) if wantarray;              # including remainder?
+    return $x;					# only x, which is [0] now
+    }
+  # check whether the numbers have the same number of elements, in that case
+  # the result will fit into one element and can be computed efficiently
+  if (@$yorg == @$x)
+    {
+    my $rem;
+    # if $yorg has more digits than $x (it's leading element is longer than
+    # the one from $x), the result will also be 0:
+    if (length(int($yorg->[-1])) > length(int($x->[-1])))
+      {
+      $rem = [@$x] if wantarray;		# make copy
+      splice (@$x,1);				# keep ref to org array
+      $x->[0] = 0;				# set to 0
+      return ($x,$rem) if wantarray;		# including remainder?
+      return $x;
+      }
+    # now calculate $x / $yorg
+    if (length(int($yorg->[-1])) == length(int($x->[-1])))
+      {
+      # same length, so make full compare, and if equal, return 1
+      # hm, same lengths, but same contents? So we need to check all parts:
+      my $a = 0; my $j = scalar @$x - 1;
+      # manual way (abort if unequal, good for early ne)
+      while ($j >= 0)
+        {
+        last if ($a = $x->[$j] - $yorg->[$j]); $j--;
+        }
+      # $a contains the result of the compare between X and Y
+      # a < 0: x < y, a == 0 => x == y, a > 0: x > y
+      if ($a <= 0)
+        {
+        if (wantarray)
+	  {
+          $rem = [ 0 ];			# a = 0 => x == y => rem 1
+          $rem = [@$x] if $a != 0;	# a < 0 => x < y => rem = x
+	  }
+        splice(@$x,1);			# keep single element
+        $x->[0] = 0;			# if $a < 0
+        if ($a == 0)
+          {
+          # $x == $y
+          $x->[0] = 1;
+          }
+        return ($x,$rem) if wantarray;
+        return $x;
+        }
+      # $x >= $y, proceed normally
+      }
+    }
+
+  # all other cases:
+
   my $y = [ @$yorg ];				# always make copy to preserve
 
   my ($car,$bar,$prd,$dd,$xi,$yi,@q,$v2,$v1,@d,$tmp,$q,$u2,$u1,$u0);
@@ -580,7 +658,7 @@ sub _div_use_mul
     $u2 = 0 unless $u2;
     #warn "oups v1 is 0, u0: $u0 $y->[-2] $y->[-1] l ",scalar @$y,"\n"
     # if $v1 == 0;
-     $q = (($u0 == $v1) ? $MAX_VAL : int(($u0*$MBASE+$u1)/$v1));
+    $q = (($u0 == $v1) ? $MAX_VAL : int(($u0*$MBASE+$u1)/$v1));
     --$q while ($v2*$q > ($u0*$MBASE+$u1-$q*$v1)*$MBASE+$u2);
     if ($q)
       {
@@ -597,11 +675,12 @@ sub _div_use_mul
 	for ($yi = 0, $xi = $#$x-$#$y-1; $yi <= $#$y; ++$yi,++$xi) 
           {
 	  $x->[$xi] -= $MBASE
-	   if ($car = (($x->[$xi] += $y->[$yi] + $car) > $MBASE));
+	   if ($car = (($x->[$xi] += $y->[$yi] + $car) >= $MBASE));
 	  }
 	}   
       }
-      pop(@$x); unshift(@q, $q);
+    pop(@$x);
+    unshift(@q, $q);
     }
   if (wantarray) 
     {
@@ -688,7 +767,7 @@ sub _div_use_div
     splice (@$x,1);				# keep ref to original array
     $x->[0] = 0;				# set to 0
     return ($x,$rem) if wantarray;		# including remainder?
-    return $x;
+    return $x;					# only x, which is [0] now
     }
   # check whether the numbers have the same number of elements, in that case
   # the result will fit into one element and can be computed efficiently
@@ -706,21 +785,27 @@ sub _div_use_div
       return $x;
       }
     # now calculate $x / $yorg
+
     if (length(int($yorg->[-1])) == length(int($x->[-1])))
       {
       # same length, so make full compare, and if equal, return 1
-      # hm, same lengths,  but same contents? So we need to check all parts:
+      # hm, same lengths, but same contents? So we need to check all parts:
       my $a = 0; my $j = scalar @$x - 1;
       # manual way (abort if unequal, good for early ne)
       while ($j >= 0)
         {
         last if ($a = $x->[$j] - $yorg->[$j]); $j--;
         }
+      # $a contains the result of the compare between X and Y
       # a < 0: x < y, a == 0 => x == y, a > 0: x > y
       if ($a <= 0)
         {
-        $rem = [@$x] if wantarray;
-        splice(@$x,1);
+        if (wantarray)
+	  {
+          $rem = [ 0 ];			# a = 0 => x == y => rem 1
+          $rem = [@$x] if $a != 0;	# a < 0 => x < y => rem = x
+	  }
+        splice(@$x,1);			# keep single element
         $x->[0] = 0;			# if $a < 0
         if ($a == 0)
           {
@@ -730,9 +815,8 @@ sub _div_use_div
         return ($x,$rem) if wantarray;
         return $x;
         }
-      # $x >= $y, proceed normally
+      # $x >= $y, so proceed normally
       }
-
     }
 
   # all other cases:
@@ -760,6 +844,10 @@ sub _div_use_div
     {
     push(@$x, 0);
     }
+
+  # @q will accumulate the final result, $q contains the current computed
+  # part of the final result
+
   @q = (); ($v2,$v1) = @$y[-2,-1];
   $v2 = 0 unless $v2;
   while ($#$x > $#$y) 
@@ -768,7 +856,7 @@ sub _div_use_div
     $u2 = 0 unless $u2;
     #warn "oups v1 is 0, u0: $u0 $y->[-2] $y->[-1] l ",scalar @$y,"\n"
     # if $v1 == 0;
-     $q = (($u0 == $v1) ? $MAX_VAL : int(($u0*$MBASE+$u1)/$v1));
+    $q = (($u0 == $v1) ? $MAX_VAL : int(($u0*$MBASE+$u1)/$v1));
     --$q while ($v2*$q > ($u0*$MBASE+$u1-$q*$v1)*$MBASE+$u2);
     if ($q)
       {
@@ -785,7 +873,7 @@ sub _div_use_div
 	for ($yi = 0, $xi = $#$x-$#$y-1; $yi <= $#$y; ++$yi,++$xi) 
           {
 	  $x->[$xi] -= $MBASE
-	   if ($car = (($x->[$xi] += $y->[$yi] + $car) > $MBASE));
+	   if ($car = (($x->[$xi] += $y->[$yi] + $car) >= $MBASE));
 	  }
 	}   
       }
@@ -827,31 +915,28 @@ sub _acmp
   # internal absolute post-normalized compare (ignore signs)
   # ref to array, ref to array, return <0, 0, >0
   # arrays must have at least one entry; this is not checked for
-
   my ($c,$cx,$cy) = @_;
+ 
+  # shortcut for short numbers 
+  return (($cx->[0] <=> $cy->[0]) <=> 0) 
+   if scalar @$cx == scalar @$cy && scalar @$cx == 1;
 
   # fast comp based on number of array elements (aka pseudo-length)
-  my $lxy = scalar @$cx - scalar @$cy;
+  my $lxy = (scalar @$cx - scalar @$cy)
+  # or length of first element if same number of elements (aka difference 0)
+    ||
+  # need int() here because sometimes the last element is '00018' vs '18'
+   (length(int($cx->[-1])) - length(int($cy->[-1])));
   return -1 if $lxy < 0;				# already differs, ret
   return 1 if $lxy > 0;					# ditto
 
-  # now calculate length based on digits, not parts
-  # we need only the length of the last element, since both array have the
-  # same number of parts
-  $lxy = length(int($cx->[-1])) - length(int($cy->[-1]));
-  return -1 if $lxy < 0;
-  return 1 if $lxy > 0;
-
-  # hm, same lengths,  but same contents? So we need to check all parts:
-  my $a; my $j = scalar @$cx - 1;
   # manual way (abort if unequal, good for early ne)
-  while ($j >= 0)
+  my $a; my $j = scalar @$cx;
+  while (--$j >= 0)
     {
-    last if ($a = $cx->[$j] - $cy->[$j]); $j--;
+    last if ($a = $cx->[$j] - $cy->[$j]);
     }
-  return 1 if $a > 0;
-  return -1 if $a < 0;
-  0;						# numbers are equal
+  $a <=> 0;
   }
 
 sub _len
@@ -1013,6 +1098,7 @@ sub _mod
     my ($xo,$rem) = _div($c,$x,$yo);
     return $rem;
     }
+
   my $y = $yo->[0];
   # both are single element arrays
   if (scalar @$x == 1)
@@ -1021,7 +1107,7 @@ sub _mod
     return $x;
     }
 
-  # @y is single element, but @x has more than one
+  # @y is a single element, but @x has more than one element
   my $b = $BASE % $y;
   if ($b == 0)
     {
@@ -1059,7 +1145,7 @@ sub _mod
     $r = 0 if $r == $y;
     $x->[0] = $r;
     }
-  splice (@$x,1);
+  splice (@$x,1);		# keep one element of $x
   $x;
   }
 
@@ -1176,63 +1262,137 @@ sub _fac
 
   if ((@$cx == 1) && ($cx->[0] <= 2))
     {
-    $cx->[0] = 1 * ($cx->[0]||1); # 0,1 => 1, 2 => 2
+    $cx->[0] ||= 1;		# 0 => 1, 1 => 1, 2 => 2
     return $cx;
     }
 
   # go forward until $base is exceeded
-  # limit is either $x or $base (x == 100 means as result too high)
+  # limit is either $x steps (steps == 100 means a result always too high) or
+  # $base.
   my $steps = 100; $steps = $cx->[0] if @$cx == 1;
-  my $r = 2; my $cf = 3; my $step = 1; my $last = $r;
-  while ($r < $BASE && $step < $steps)
+  my $r = 2; my $cf = 3; my $step = 2; my $last = $r;
+  while ($r*$cf < $BASE && $step < $steps)
     {
     $last = $r; $r *= $cf++; $step++;
     }
-  if ((@$cx == 1) && ($step == $cx->[0]))
+  if ((@$cx == 1) && $step == $cx->[0])
     {
-    # completely done
-    $cx = [$last];
+    # completely done, so keep reference to $x and return
+    $cx->[0] = $r;
     return $cx;
     }
+  
   # now we must do the left over steps
-
-  # do so as long as n has more than one element
-  my $n = $cx->[0];
-  # as soon as the last element of $cx is 0, we split it up and remember how
-  # many zeors we got so far. The reason is that n! will accumulate zeros at
-  # the end rather fast.
-  my $zero_elements = 0;
-  $cx = [$last];
+  my $n;					# steps still to do
   if (scalar @$cx == 1)
     {
-    my $n = _copy($c,$cx);
-    # no need to test for $steps, since $steps is a scalar and we stop before
-    while (scalar @$n != 1)
+    $n = $cx->[0];
+    }
+  else
+    {
+    $n = _copy($c,$cx);
+    }
+
+  $cx->[0] = $last; splice (@$cx,1);		# keep ref to $x
+  my $zero_elements = 0;
+
+  # do left-over steps fit into a scalar?
+  if (ref $n eq 'ARRAY')
+    {
+    # No, so use slower inc() & cmp()
+    $step = [$step];
+    while (_acmp($step,$n) <= 0)
       {
+      # as soon as the last element of $cx is 0, we split it up and remember
+      # how many zeors we got so far. The reason is that n! will accumulate
+      # zeros at the end rather fast.
       if ($cx->[0] == 0)
         {
         $zero_elements ++; shift @$cx;
         }
-      _mul($c,$cx,$n); _dec($c,$n);
+      _mul($c,$cx,$step); _inc($c,$step);
       }
-    $n = $n->[0];		# "convert" to scalar
     }
-  
-  # the left over steps will fit into a scalar, so we can speed it up
-  while ($n != $step)
+  else
     {
-    if ($cx->[0] == 0)
+    # Yes, so we can speed it up slightly
+    while ($step <= $n)
       {
-      $zero_elements ++; shift @$cx;
+      # When the last element of $cx is 0, we split it up and remember
+      # how many we got so far. The reason is that n! will accumulate
+      # zeros at the end rather fast.
+      if ($cx->[0] == 0)
+        {
+        $zero_elements ++; shift @$cx;
+        }
+      _mul($c,$cx,[$step]); $step++;
       }
-    _mul($c,$cx,[$n]); $n--;
     }
   # multiply in the zeros again
   while ($zero_elements-- > 0)
     {
     unshift @$cx, 0; 
     }
-  $cx;
+  $cx;			# return result
+  }
+
+sub _log_int
+  {
+  # calculate integer log of $x to base $base
+  # ref to array, ref to array - return ref to array
+  my ($c,$x,$base) = @_;
+
+  # X == 0 => NaN
+  return if (scalar @$x == 1 && $x->[0] == 0);
+  # BASE 0 or 1 => NaN
+  return if (scalar @$base == 1 && $base->[0] < 2);
+  my $cmp = _acmp($c,$x,$base);
+  # X == BASE => 1
+  if ($cmp == 0)
+    {
+    splice (@$x,1); $x->[0] = 1;
+    return ($x,1)
+    }
+  # X < BASE
+  if ($cmp < 0)
+    {
+    splice (@$x,1); $x->[0] = 0;
+    return ($x,undef);
+    }
+
+  # this trial multiplication is very fast, even for large counts (like for
+  # 2 ** 1024, since this still requires only 1024 very fast steps
+  # (multiplication of a large number by a very small number is very fast))
+  my $x_org = _copy($c,$x);		# preserve x
+  splice(@$x,1); $x->[0] = 1;		# keep ref to $x
+
+  # simple loop that increments $x by two in each step, possible overstepping
+  # the real result by one
+
+  # use a loop that keeps $x as scalar as long as possible (this is faster)
+  my $trial = _copy($c,$base); my $a;
+  my $base_mul = _mul($c, _copy($c,$base), $base);
+
+  while (($a = _acmp($x,$trial,$x_org)) < 0)
+    {
+    _mul($c,$trial,$base_mul); _add($c, $x, [2]);
+    }
+
+  my $exact = 1;
+  if ($a > 0)
+    {
+    # overstepped the result
+    _dec($c, $x);
+    _div($c,$trial,$base);
+    $a = _acmp($x,$trial,$x_org);
+    if ($a > 0)
+      {
+      _dec($c, $x);
+      }
+    $exact = 0 if $a != 0;
+    }
+  
+  ($x,$exact);				# return result
   }
 
 # for debugging:
@@ -1321,8 +1481,6 @@ sub _sqrt
 sub _root
   {
   # take n'th root of $x in place (n >= 3)
-  # Compute a guess of the result (by rule of thumb), then improve it via
-  # Newton's method.
   my ($c,$x,$n) = @_;
  
   if (scalar @$x == 1)
@@ -1335,13 +1493,80 @@ sub _root
     else
       {
       # fit's into one Perl scalar, so result can be computed directly
-      $x->[0] = int( $x->[0] ** (1 / $n->[0]) );
+      # cannot use int() here, because it rounds wrongly (try 
+      # (81 ** 3) ** (1/3) to see what I mean)
+      #$x->[0] = int( $x->[0] ** (1 / $n->[0]) );
+      # round to 8 digits, then truncate result to integer
+      $x->[0] = int ( sprintf ("%.8f", $x->[0] ** (1 / $n->[0]) ) );
       }
     return $x;
     } 
 
-  # XXX TODO
+  # X is more than one element
+  # if $n is a power of two, we can repeatedly take sqrt($X) and find the
+  # proper result, because sqrt(sqrt($x)) == root($x,4)
+  my $b = _as_bin($c,$n);
+  if ($$b =~ /0b1(0+)/)
+    {
+    my $count = CORE::length($1);	# 0b100 => len('00') => 2
+    my $cnt = $count;			# counter for loop
+    unshift (@$x, 0);			# add one element, together with one
+					# more below in the loop this makes 2
+    while ($cnt-- > 0)
+      {
+      # 'inflate' $X by adding one element, basically computing
+      # $x * $BASE * $BASE. This gives us more $BASE_LEN digits for result
+      # since len(sqrt($X)) approx == len($x) / 2.
+      unshift (@$x, 0);
+      # calculate sqrt($x), $x is now one element to big, again. In the next
+      # round we make that two, again.
+      _sqrt($c,$x);
+      }
+    # $x is now one element to big, so truncate result by removing it
+    splice (@$x,0,1);
+    } 
+  else
+    {
+    # trial computation by starting with 2,4,8,16 etc until we overstep
 
+    my $step = _two();
+    my $trial = _two();
+
+    _mul($c, $trial, $step) 
+      while (_acmp($c, _pow($c, _copy($c, $trial), $n), $x) < 0);
+
+    # hit exactly?
+    if (_acmp($c, _pow($c, _copy($c, $trial), $n), $x) == 0)
+      {
+      @$x = @$trial;			# make copy while preserving ref to $x
+      return $x;
+      }
+    # overstepped, so go back on step
+    _div($c, $trial, $step);
+
+    # add two, because $trial cannot be exactly the result (otherwise we would
+    # alrady have found it)
+    _add($c, $trial, $step);
+ 
+    # and now add more and more (2,4,6,8, etc)
+    _add($c, $trial, $step) 
+      while (_acmp($c, _pow($c, _copy($c, $trial), $n), $x) < 0);
+
+    # hit not exactly? (overstepped)
+    # 80 too small, 81 slightly too big, 82 too big
+    if (_acmp($c, _pow($c, _copy($c, $trial), $n), $x) > 0)
+      {
+      _dec($c,$trial);
+      }
+    # 80 too small, 81 slightly too big
+    if (_acmp($c, _pow($c, _copy($c, $trial), $n), $x) > 0)
+      {
+      _dec($c,$trial);
+      }
+    
+    @$x = @$trial;			# make copy while preserving ref to $x
+    return $x;
+    }
   $x; 
   }
 
@@ -1463,10 +1688,10 @@ sub _as_hex
   # convert a decimal number to hex (ref to array, return ref to string)
   my ($c,$x) = @_;
 
-  # fit's into one element
+  # fit's into one element (handle also 0x0 case)
   if (@$x == 1)
     {
-    my $t = '0x' . sprintf("%x",$x->[0]);
+    my $t = sprintf("0x%x",$x->[0]);
     return \$t;
     }
 
@@ -1482,10 +1707,11 @@ sub _as_hex
     {
     $x10000 = [ 0x1000 ]; $h = 'h3';
     }
-  while (! _is_zero($c,$x1))
+  # while (! _is_zero($c,$x1))
+  while (@$x1 != 1 || $x1->[0] != 0)		# _is_zero()
     {
     ($x1, $xr) = _div($c,$x1,$x10000);
-    $es .= unpack($h,pack('v',$xr->[0]));
+    $es .= unpack($h,pack('v',$xr->[0]));	# XXX TODO: why pack('v',...)?
     }
   $es = reverse $es;
   $es =~ s/^[0]+//;   # strip leading zeros
@@ -1498,10 +1724,15 @@ sub _as_bin
   # convert a decimal number to bin (ref to array, return ref to string)
   my ($c,$x) = @_;
 
-  # fit's into one element
-  if (@$x == 1)
+  # fit's into one element (and Perl recent enough), handle also 0b0 case
+  # handle zero case for older Perls
+  if ($] <= 5.005 && @$x == 1 && $x->[0] == 0)
     {
-    my $t = '0b' . sprintf("%b",$x->[0]);
+    my $t = '0b0'; return \$t;
+    }
+  if (@$x == 1 && $] >= 5.006)
+    {
+    my $t = sprintf("0b%b",$x->[0]);
     return \$t;
     }
   my $x1 = _copy($c,$x);
@@ -1516,10 +1747,12 @@ sub _as_bin
     {
     $x10000 = [ 0x1000 ]; $b = 'b12';
     }
-  while (! _is_zero($c,$x1))
+  # while (! _is_zero($c,$x1))
+  while (!(@$x1 == 1 && $x1->[0] == 0))		# _is_zero()
     {
     ($x1, $xr) = _div($c,$x1,$x10000);
-    $es .= unpack($b,pack('v',$xr->[0]));
+    $es .= unpack($b,pack('v',$xr->[0]));	# XXX TODO: why pack('v',...)?
+    # $es .= unpack($b,$xr->[0]);
     }
   $es = reverse $es;
   $es =~ s/^[0]+//;   # strip leading zeros
@@ -1556,7 +1789,7 @@ sub _from_bin
   # convert a hex number to decimal (ref to string, return ref to array)
   my ($c,$bs) = @_;
 
-  # instead of converting 8 bit at a time, it is faster to convert the
+  # instead of converting X (8) bit at a time, it is faster to "convert" the
   # number to hex, and then call _from_hex.
 
   my $hs = $$bs;
@@ -1564,27 +1797,8 @@ sub _from_bin
   my $l = length($hs);					# bits
   $hs = '0' x (8-($l % 8)) . $hs if ($l % 8) != 0;	# padd left side w/ 0
   my $h = unpack('H*', pack ('B*', $hs));		# repack as hex
-  return $c->_from_hex(\('0x'.$h));
- 
-  my $mul = _one();
-  my $m = [ 0x100 ];				# 8 bit at a time
-  my $x = _zero();
-
-  my $len = length($$bs)-2;
-  $len = int($len/8);				# 4-digit parts, w/o '0x'
-  my $val; my $i = -8;
-  while ($len >= 0)
-    {
-    $val = substr($$bs,$i,8);
-    $val =~ s/^[+-]?0b// if $len == 0;		# for last part only
-
-    $val = ord(pack('B8',substr('00000000'.$val,-8,8))); 
-
-    $i -= 8; $len --;
-    _add ($c, $x, _mul ($c, [ $val ], $mul ) ) if $val != 0;
-    _mul ($c, $mul, $m ) if $len >= 0; 		# skip last mul
-    }
-  $x;
+  
+  $c->_from_hex(\('0x'.$h));
   }
 
 ##############################################################################
@@ -1671,8 +1885,8 @@ Math::BigInt::Calc - Pure Perl module to support Math::BigInt
 =head1 SYNOPSIS
 
 Provides support for big integer calculations. Not intended to be used by other
-modules (except Math::BigInt::Cached). Other modules which sport the same
-functions can also be used to support Math::BigInt, like Math::BigInt::Pari.
+modules. Other modules which sport the same functions can also be used to support
+Math::BigInt, like Math::BigInt::GMP or Math::BigInt::Pari.
 
 =head1 DESCRIPTION
 
@@ -1758,6 +1972,10 @@ slow) fallback routines to emulate these:
 	_and(obj1,obj2)	AND (bit-wise) object 1 with object 2
 	_or(obj1,obj2)	OR (bit-wise) object 1 with object 2
 
+	_signed_or
+	_signed_and
+	_signed_xor
+
 	_mod(obj,obj)	Return remainder of div of the 1st by the 2nd object
 	_sqrt(obj)	return the square root of object (truncated to int)
 	_root(obj)	return the n'th (n >= 3) root of obj (truncated to int)
@@ -1768,6 +1986,12 @@ slow) fallback routines to emulate these:
 	_zeros(obj)	return number of trailing decimal zeros
 	_modinv		return inverse modulus
 	_modpow		return modulus of power ($x ** $y) % $z
+	_log_int(X,N)	calculate integer log() of X in base N
+			X >= 0, N >= 0 (return undef for NaN)
+			returns (RESULT, EXACT) where EXACT is:
+			 1     : result is exactly RESULT
+			 0     : result was truncated to RESULT
+			 undef : unknown whether result is exactly RESULT
 
 Input strings come in as unsigned but with prefix (i.e. as '123', '0xabc'
 or '0b1101').
@@ -1815,7 +2039,7 @@ the same terms as Perl itself.
 Original math code by Mark Biggar, rewritten by Tels L<http://bloodgate.com/>
 in late 2000.
 Seperated from BigInt and shaped API with the help of John Peacock.
-Fixed/enhanced by Tels 2001-2002.
+Fixed, sped-up and enhanced by Tels http://bloodgate.com 2001-2003.
 
 =head1 SEE ALSO
 
